﻿using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using RhythmBullet.Zer01HD.ContentResolvers;
using RhythmBullet.Zer01HD.Preferences;
using RhythmBullet.Zer01HD.Screens.MainMenu;
using RhythmBullet.Zer01HD.Utilities;
using RhythmBullet.Zer01HD.Utilities.Camera;
using RhythmBullet.Zer01HD.Utilities.ContentSystem;
using RhythmBullet.Zer01HD.Utilities.DataTypes;
using RhythmBullet.Zer01HD.Utilities.Input;
using RhythmBullet.Zer01HD.Utilities.Persistence;
using RhythmBullet.Zer01HD.Utilities.ScreenSystem;
using System;
using System.Diagnostics;

namespace RhythmBullet
{
    /// <summary>
    /// This is the main type for your game.
    /// </summary>
    public class RhythmBulletGame : Game
    {
        public const int WORLD_WIDTH = 4;
        public const int WORLD_HEIGHT = 3;
        public static int pixels_per_WU;

        GraphicsDeviceManager graphics;
        SpriteBatch spriteBatch;
        public readonly ContentManagerController Assets;
        readonly ResolutionContentResolver resolutionContentResolver;
        public PreferencesManager preferencesManager;
        private ScreenManager screenManager;
        public Camera2D Camera { get; private set; }
        private bool resizing;
        private bool initialLoadComplete;
        private MainScreen mainScreen;
        private Texture2D currentCursorTexture;

        public RhythmBulletGame()
        {
            graphics = new GraphicsDeviceManager(this);
            Content.RootDirectory = "Content";
            Assets = new ContentManagerController(Content);

            resolutionContentResolver = new ResolutionContentResolver();
            FontContentResolver fcr = new FontContentResolver(resolutionContentResolver);
            Assets.contentPathModifier.Add(typeof(Texture2D), resolutionContentResolver);
            Assets.contentPathModifier.Add(typeof(SpriteFont), fcr);
            preferencesManager = new PreferencesManager(Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData) + "/RhythmBullet",
                new General(),
                new Controls());
            preferencesManager.Load();
            Resolution resolution = preferencesManager.GetPreferences<General>().Resolution;
            graphics.PreferredBackBufferWidth = resolution.Width;
            graphics.PreferredBackBufferHeight = resolution.Height;
        }

        /// <summary>
        /// Allows the game to perform any initialization it needs to before starting to run.
        /// This is where it can query for any required services and load any non-graphic
        /// related content. Calling base.Initialize will enumerate through any components
        /// and initialize them as well.
        /// </summary>
        protected override void Initialize()
        {
            Resolution resolution = preferencesManager.GetPreferences<General>().Resolution;
            resolutionContentResolver.Width = resolution.Width;
            resolutionContentResolver.Height = resolution.Height;
            Debug.WriteLine("Initial setup complete.");
            base.Initialize();
        }

        /// <summary>
        /// LoadContent will be called once per game and is the place to load
        /// all of your content.
        /// </summary>
        protected override void LoadContent()
        {
            // Create a new SpriteBatch, which can be used to draw textures.
            spriteBatch = new SpriteBatch(GraphicsDevice);
            Camera = new Camera2D(graphics.GraphicsDevice);
            screenManager = new ScreenManager(graphics, Camera);
            screenManager.RequireNextScreenEvent += RequireNextScreen;
            QueueContent();
            screenManager.Screen = new LoadingScreen(this, Content.Load<Texture2D>("RhythmBullet"), 0.7f);
        }

        /// <summary>
        /// UnloadContent will be called once per game and is the place to unload
        /// game-specific content.
        /// </summary>
        protected override void UnloadContent()
        {
            Assets.UnloadAll();
            screenManager.Dispose();
            currentCursorTexture.Dispose();
        }

        /// <summary>
        /// Allows the game to run logic such as updating the world,
        /// checking for collisions, gathering input, and playing audio.   
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Update(GameTime gameTime)
        {
            if (!Assets.Done)
            {
                Assets.Update();
            }
            else if (resizing)
            {
                resizing = false;
                PostResize();
            }
            else if (!initialLoadComplete)
            {
                Debug.WriteLine("Initial load complete.");
                initialLoadComplete = true;
                SetUpCursor();
                mainScreen = new MainScreen(Assets);
            }

            screenManager.UpdateCurrentScreen(gameTime, Assets.Done);

            InputUtilities.Update();
            Camera.Update();
            base.Update(gameTime);
        }

        /// <summary>
        /// This is called when the game should draw itself.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Draw(GameTime gameTime)
        {
            screenManager.DrawCurrentScreen(spriteBatch);
            base.Draw(gameTime);
        }

        public void Resize(int width, int height)
        {
            pixels_per_WU = height / WORLD_HEIGHT;
            resolutionContentResolver.Width = width;
            resolutionContentResolver.Height = height;
            graphics.PreferredBackBufferWidth = width;
            graphics.PreferredBackBufferHeight = height;
            resizing = true;
            Assets.UnloadAll();
            QueueContent();
            screenManager.Resize(new LoadingScreen(this, Content.Load<Texture2D>("loading_ring"), 0.4f));
            SetUpCursor();
        }

        private void PostResize()
        {
            graphics.ApplyChanges();
            screenManager.PostResize();
        }

        private void QueueContent()
        {
            Assets.Queue<Texture2D>("cursor", false);
            Assets.Queue<Texture2D>("title");
            Assets.Queue<Texture2D>("default_cover", false);
            Assets.Queue<Texture2D>("backgrounds/mainBG");
        }

        private void SetUpCursor()
        {
            Texture2D texture = Assets.Get<Texture2D>("cursor");
            int cursorSize = (int)(0.08f * graphics.PreferredBackBufferHeight);
            Debug.WriteLine("Cursor size length: " + cursorSize);
            RenderTarget2D renderTarget = new RenderTarget2D(GraphicsDevice, cursorSize, cursorSize);
            GraphicsDevice.SetRenderTarget(renderTarget);
            GraphicsDevice.Clear(Color.Transparent);
            spriteBatch.Begin();
            spriteBatch.Draw(texture, new Rectangle(0, 0, cursorSize, cursorSize), Color.White);
            spriteBatch.End();
            GraphicsDevice.SetRenderTarget(null);

            currentCursorTexture?.Dispose();
            currentCursorTexture = renderTarget;
            Mouse.SetCursor(MouseCursor.FromTexture2D(currentCursorTexture, currentCursorTexture.Width / 2, currentCursorTexture.Height / 2));
        }

        private void RequireNextScreen(Screen Screen)
        {
            Screen.NextScreen = mainScreen;
        }
    }
}
